#define TEST_THREE

#include <iostream>

#if defined(TEST_ONE)
#include "Delegate.h"
using namespace util;
#define make_delegate UTIL_MAKE_DELEGATE
#define delegate Delegate
#elif defined(TEST_TWO)
#include "fun_ptr2.h"
using namespace util2;
#elif defined(TEST_THREE)
#include "fun_ptr3.h"
#include "event.h"
using namespace util3;
#else
#error "must define test"
#endif

#include <iostream>
#include <cassert>

struct Foo {
   int testFn(int a)
   {
      return a + 1;
   }

   int testFnConst(int a) const
   {
      return a + 4;
   }
};

struct Bar {
   int testFn(int a)
   {
      return a + 2;
   }

   int testFnConst(int a) const
   {
      return a + 5;
   }
};

int testFnGlobal(int a)
{
   return a + 25;
}

class High_score
{
public:
   //a public event
   util::event<void(int)> newHighScoreEvent;

   High_score()
      : newHighScoreEvent()
      , newHighScoreEventInvoker(newHighScoreEvent.get_invoker())
      , high_score(0)
   {
   }

   void submit_score(int score)
   {
      if (score > high_score) {
         high_score = score;
         newHighScoreEventInvoker(high_score);
      }
   }

private:
   //a private invoker for the event
   util::event<void(int)>::invoker_type newHighScoreEventInvoker;

   int high_score;
};

void print_new_high_score(int newHighScore)
{
   std::cout << "new high score = " << newHighScore << "\n";
}

int main()
{
   std::printf("---------------------------------\n");

   Foo foo;
   Bar bar;

   {
      auto d1 = make_fun(&Foo::testFn, &foo);
      assert(d1); // operator bool()

      assert(d1 != nullptr);
      assert(!(d1 == nullptr));
      assert(nullptr != d1);
      assert(!(nullptr == d1));

      assert(d1 == d1);    // operator==
      assert(!(d1 != d1)); // operator!=
      assert(d1(1) == 2);  // operator()

      auto d2 = d1; // copy constructor
      assert(d2);   // explicit operator bool()
      assert(d2 != nullptr);
      assert(!(d2 == nullptr));
      assert(nullptr != d2);
      assert(!(nullptr == d2));
      assert(d2 == d2);    // operator==
      assert(d1 == d2);    // operator==
      assert(d2 == d1);    // operator==
      assert(!(d2 != d2)); // operator==
      assert(!(d1 != d2)); // operator==
      assert(!(d2 != d1)); // operator==

      d2 = nullptr; // operator=(std::nullptr_t);
      assert(!d2);  // explicit operator bool()

      assert(d2 == nullptr);
      assert(!(d2 != nullptr));
      assert(nullptr == d2);
      assert(!(nullptr != d2));

      assert(d1 != d2);    // operator!=
      assert(!(d1 == d2)); // operator!=

      assert(d2 != d1);    // operator!=
      assert(!(d2 == d1)); // operator!=

      assert(d2 == d2);    // operator==
      assert(!(d2 != d2)); // operator==

      d2 = make_fun(&Bar::testFn, &bar); // operator=(const delegate& rhs)
      assert(d2);                        // explicit operator bool()

      assert(d2 != nullptr);
      assert(!(d2 == nullptr));
      assert(nullptr != d2);
      assert(!(nullptr == d2));

      assert(d1 != d2);    // operator!=
      assert(!(d1 == d2)); // operator!=

      assert(d2 != d1);    // operator!=
      assert(!(d2 == d1)); // operator!=

      assert(d2 == d2);    // operator==
      assert(!(d2 != d2)); // operator==

      assert(d2(1) == 3); // operator()

      auto d3 = fun_ptr<int(int)>(nullptr); // delegate(std::nullptr_t);
      assert(!d3);                          // explicit operator bool()

      assert(d3 == nullptr);
      assert(!(d3 != nullptr));
      assert(nullptr == d3);
      assert(!(nullptr != d3));

      assert(d3 == d3);    // operator==
      assert(!(d3 != d3)); // operator!=

      assert(d3 != d2);    // operator!=
      assert(!(d3 == d2)); // operator==

      assert(d3 != d1);    // operator!=
      assert(!(d3 == d1)); // operator==

      auto d4 = fun_ptr<int(int)>(); // delegate();
      assert(!d4);                   // explicit operator bool()

      assert(d4 == nullptr);
      assert(!(d4 != nullptr));
      assert(nullptr == d4);
      assert(!(nullptr != d4));

      assert(d4 != d2);    // operator!=
      assert(!(d4 == d2)); // operator==

      assert(d4 != d1);    // operator!=
      assert(!(d4 == d1)); // operator==

      assert(d4 == d4);    // operator==
      assert(!(d4 != d4)); // operator!=

      assert(d4 == d3);    // operator==
      assert(!(d4 != d3)); // operator!=

      d1 = nullptr;
      assert(d1 == d4);
      assert(!(d1 != d4));

      d2 = nullptr;
      assert(d1 == d2);
      assert(!(d1 != d2));
   }

   {
      auto d1 = make_fun(&Foo::testFnConst, &foo);
      assert(d1); // operator bool()

      assert(d1 != nullptr);
      assert(!(d1 == nullptr));
      assert(nullptr != d1);
      assert(!(nullptr == d1));

      assert(d1 == d1);    // operator==
      assert(!(d1 != d1)); // operator!=

      assert(d1(1) == 5); // operator()

      auto d2 = d1; // copy constructor
      assert(d2);   // explicit operator bool()

      assert(d2 != nullptr);
      assert(!(d2 == nullptr));
      assert(nullptr != d2);
      assert(!(nullptr == d2));

      assert(d2 == d2);    // operator==
      assert(!(d2 != d2)); // operator==

      assert(d1 == d2);    // operator==
      assert(!(d1 != d2)); // operator==

      assert(d2 == d1);    // operator==
      assert(!(d2 != d1)); // operator==

      d2 = nullptr; // operator=(std::nullptr_t);
      assert(!d2);  // explicit operator bool()

      assert(d2 == nullptr);
      assert(!(d2 != nullptr));
      assert(nullptr == d2);
      assert(!(nullptr != d2));

      assert(d1 != d2);    // operator!=
      assert(!(d1 == d2)); // operator!=

      assert(d2 != d1);    // operator!=
      assert(!(d2 == d1)); // operator!=

      assert(d2 == d2);    // operator==
      assert(!(d2 != d2)); // operator==

      d2 = make_fun(&Foo::testFn, &foo); // operator=(const delegate& rhs)
      assert(d2);                        // explicit operator bool()

      assert(d2 != nullptr);
      assert(!(d2 == nullptr));
      assert(nullptr != d2);
      assert(!(nullptr == d2));

      assert(d1 != d2);    // operator!=
      assert(!(d1 == d2)); // operator!=

      assert(d2 != d1);    // operator!=
      assert(!(d2 == d1)); // operator!=

      assert(d2 == d2);    // operator==
      assert(!(d2 != d2)); // operator==

      assert(d2(1) == 2); // operator()

      d2 = make_fun(&Foo::testFnConst, &foo);
      assert(d2); // explicit operator bool()

      assert(d2 != nullptr);
      assert(!(d2 == nullptr));
      assert(nullptr != d2);
      assert(!(nullptr == d2));

      assert(d1 == d2);    // operator!=
      assert(!(d1 != d2)); // operator!=

      assert(d2 == d1);    // operator!=
      assert(!(d2 != d1)); // operator!=

      assert(d2 == d2);    // operator==
      assert(!(d2 != d2)); // operator==

      assert(d2(1) == 5); // operator()

      auto d3 = make_fun(&testFnGlobal);
      assert(d3); // explicit operator bool()

      assert(d3 != nullptr);
      assert(!(d3 == nullptr));
      assert(nullptr != d3);
      assert(!(nullptr == d3));

      assert(d1 != d3);    // operator!=
      assert(!(d1 == d3)); // operator!=

      assert(d3 != d1);    // operator!=
      assert(!(d3 == d1)); // operator!=

      assert(d2 != d3);    // operator!=
      assert(!(d2 == d3)); // operator!=

      assert(d3 != d2);    // operator!=
      assert(!(d3 == d2)); // operator!=

      assert(d3 == d3);    // operator==
      assert(!(d3 != d3)); // operator==

      assert(d3(1) == 26); // operator()

      auto d4 = make_fun(&testFnGlobal);
      assert(d4); // explicit operator bool()

      assert(d4 != nullptr);
      assert(!(d4 == nullptr));
      assert(nullptr != d4);
      assert(!(nullptr == d4));

      assert(d3 == d4);
      assert(!(d3 != d4));
   }

   assert(std::printf("tests done\n"));
   std::printf("---------------------------------\n");

   High_score hs;

   hs.newHighScoreEvent.attach(make_fun(&print_new_high_score));

   hs.submit_score(1);
   hs.submit_score(3);
   hs.submit_score(2);

   hs.newHighScoreEvent.detach(make_fun(&print_new_high_score));

   hs.submit_score(4);

   return 0;
}
